/* Copyright (C) 2000 Stuart Mark Pomerantz
 *
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */

#include <GL/glut.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <string.h>

/* for reading textures */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

/* for threads */
#include <pthread.h>
#include <semaphore.h>

/* for gettimeofday ( microsecond clock) */
#include <sys/time.h>

/* for OSS/Free sound */
#include <sys/soundcard.h>
#include <sys/ioctl.h>

/*
 * For the copyright
 */

#define COPYRIGHT "Asteroids3D v0.2.2 Copyright (C) 2000 Stuart Mark Pomerantz\n"

/*
 * for Frustum perspective projection
 */
#define ZNEAR   1
#define ZFAR    100

#define LEFT    -1
#define RIGHT    1

#define TOP      1
#define BOTTOM  -1


/* What's your favorite kind of pi ? */
#define PI M_PI

/* for glutTimerFuncs */
#define MSEC    1

/* maximum thrust that can be input and max thrust per unit time */
#define MAX_TINPUT 0.5
#define MAX_TPS 25.0

/* maximum rotation that can be input per unit time in degrees */
#define MAX_DPS_INPUT 75.0

/*
 * Asteroid generation parameters
 */

/*
 * generate asteroids within a cubic space which is
 * no further than this distance from the origin
 */
#define AGB 50.0      /* That's asteroid generation boundary */
#define MAX_ARD 180.0   /* max asteroid rotation delta (per unit time)*/
#define MAX_APD 20.0   /* max asteroid position delta length (per unit time)*/
#define ABOUND  90  /* asteroid torus wrapping boundary */
#define NUM_AST_TYPES 6 /* number of different kinds of asteroids/display lists */
#define MIN_AST_MASS 5.0
#define AST_MASS_K 10.0

/* Torpedos are the "bullets" in the game. 
 * You can only fire a few at a time 
 */
#define MAX_TORPS 5
#define TORP_TTL 1.8
#define TORP_DELTA 25.0 
#define TORP_RADIUS 0.5
#define TORP_RADIUS_DELTA 0.2
#define TORP_BB_ZROT 5  /* the number of degrees to rotate the torp billboard about the z axis */

/* torpedoes leave trails behind them */
#define MAX_TORP_TRAILS (MAX_TORPS*3)
#define TORP_TRAIL_LEN  25
#define NO_TRAIL -1

/* space ship size and mass */
#define SHIP_SZ 3.0
#define SHIP_MASS 1.0

/* shield flash duration */
#define SHIELD_TTL 1.0 

/* number of dust particles */
#define NUM_DUST_PART 50
#define NUM_CHILD_DUST_PART 10
#define CHILD_DUST_PART_DELTA 20 /* maximum number of units a child can be from the parent dust particle */

/* for explosions */
#define MAX_BLASTS (MAX_TORPS+10)
#define BLAST_TTL 5
#define BLAST_ER 2.0 /* blast expansion rate */
#define NUM_BLAST_TYPES 5

/* nebula billboard textures */
#define NUM_NEBULA_TEX 3

/* initial shield/max shield strength, and shield recharge rate per second */
#define MAX_SHIELD 4000
#define SHIELD_REGEN_PS 10

/* When you run through an explosion the particles smack up
 * against your "windshield" pretty much just a little eye candy
 * anyway you could potentially ( though unlikely ) run through
 * MAX_BLASTS explosions. And I need to also set about how many
 * particles you impact per blast which is: 
 *
 * MAX_WP_TTL is in seconds, of course
 */
#define NUM_WP_PER_BLAST 10
#define MAX_WP_TTL 2



/* maximum time to flicker the hud gauges after a collision */
#define MAX_HUD_FLICKER_TTL 5.0


/* The code which generates data for rendering
 * will cycle some number of ticks per second.
 * 
 * On a "slow" computer, this code will go slower
 * than the number of ticks per second. The limiting
 * factor should always be the time to render, however.
 *
 * On a "fast" computer, this code will cycle no
 * faster than ticks per second, so the game should
 * never be un-playable on a "fast" computer.
 *
 * TPS = ticks per second
 */

#define TPS 60.0


/* size of the sound_buffer in, of course, bytes */
#define MAX_SOUND_BUF_BYTES  32768

/* define the types of sounds that play_sound() recognizes */
#define SOUND_ASTEROID_HIT     0
#define SOUND_TARGET_AQUIRED   1
#define SOUND_COLLISION_WARN   2
#define SOUND_COLLISION        4
#define SOUND_TORP_FIRED       5

/* This stuff is for the score file */
#define MAX_TOKEN_LEN 50
#define SCORE_FILE "/tmp/asteroids-scores"
#define NUM_SCORES 10
#define SCORE_FILE_SEPARATOR_CHAR ','
#define SCORE_FILE_COMMENT_CHAR   '#'
#define SCORE_FILE_ESCAPE_CHAR    '\\'

/* This stuff for the rc file */
#define RC_FILENAME "/.a3drc"
#define RC_FILE_SEPARATOR_CHAR '='
#define RC_FILE_COMMENT_CHAR '#'
#define RC_FILE_ESCAPE_CHAR    '\\'

/* standard hud colors */
#define HUD_R 0.3
#define HUD_G 1.0
#define HUD_B 0.3
#define HUD_A 0.7

/* lens flares */
#define MAX_FLARES 10


/* This is the default keyboard map ( see the keyboard_map struct below ) */
#define END_GAME_KEY 27   /* the escape key */
#define PAUSE_KEY 'p'

#define SLIDE_FORWARD_KEY 'r'
#define SLIDE_BACKWARD_KEY 'v'
#define SLIDE_LEFT_KEY 'd'
#define SLIDE_RIGHT_KEY 'g'
#define SLIDE_UP_KEY 'e'
#define SLIDE_DOWN_KEY  't'
#define STOP_KEY 'f'
#define ROT_ABOUTX_CW_KEY '8'
#define ROT_ABOUTX_CCW_KEY '2'
#define ROT_ABOUTY_CW_KEY '6'
#define ROT_ABOUTY_CCW_KEY '4'
#define ROT_ABOUTZ_CW_KEY '7'
#define ROT_ABOUTZ_CCW_KEY '9'

#define HEADLIGHT_TOGGLE_KEY '['
#define TARGET_ASTEROID_KEY 'b'
#define AUTO_TARGET_ASTEROID_KEY '\t'
#define FIRE_TORP_KEY ' '

#define CROSSHAIR_TOGGLE_KEY '+'
#define DRAW_TRACER_TOGGLE_KEY '_'
#define DRAW_SHIELD_TOGGLE_KEY ')'
#define DRAW_DUST_TOGGLE_KEY '('
#define DRAW_TORP_LENSFLARE_TOGGLE_KEY '*'
#define ASTEROID_TEXTURE_TOGGLE_KEY '&'
#define NEBULA_TEXTURE_TOGGLE_KEY '^'
#define TORP_TEXTURE_TOGGLE_KEY '%'
#define AXES_TOGGLE_KEY '$'
#define PLAY_SOUND_TOGGLE_KEY '#'
#define REVERSE_MOUSE_X_TOGGLE_KEY '!'
#define REVERSE_MOUSE_Y_TOGGLE_KEY '@'


/* This is the default user options map ( see the user_options_map struct below ) */
#define AXES_FLAG 0
#define CROSSHAIR_FLAG 1
#define DRAW_TRACER_FLAG 1
#define DRAW_SHIELD_FLAG 1
#define DRAW_DUST_FLAG 1
#define TEXTURE_FLAG 0
#define NEBULA_TEXTURE_FLAG 0
#define PLAY_SOUND_FLAG 1
#define TORP_TEXTURE_FLAG 0
#define DRAW_TORP_LENSFLARE_FLAG 0

#define MOUSE_XDIR 1 /* 1 or -1 */
#define MOUSE_YDIR 1 /* 1 or -1 */
#define NUM_ASTEROIDS 100 /* initial number of asteroids 150 is good */
#define WINDOWX 800 /* window size in pixels */
#define WINDOWY 800
#define WINDOW_XPOS 150 /* window position offset */
#define WINDOW_YPOS 50

/* these are keyboar rotation scaling factors notice that if you multiply
 * KEYB_BASE_ROT_SCALE by KEYB_MAX_ROT_SCALE you get 1
 */

#define KEYB_BASE_ROT_SCALE 0.05
#define KEYB_MAX_ROT_SCALE  20
#define KEYB_SCALE_DECAY 2

/*
 * structures / typedefs
 */
 
struct lens_flare {
	int   dl ;
	float pos_scalar ;
	float size_scalar ;
	float r,g,b,a ;
	float delta_z ;
} ;

typedef struct lens_flare lens_flare ;


struct sound {
	char *data ;
	int  size ;
} ;

typedef struct sound sound ;

 
struct vector {
	float x ;
	float y ;
	float z ;
	float w ;
} ;

typedef struct vector vector ;



struct asteroid {
	float  mass ;          /* BASED ON TYPE */
	int	   type ;          /* which asteroid display list to use */
	float  radius ;		   /* radius to use for collisions -- BASED ON TYPE */
	vector pos ;           /* current position */
	vector velocity ;      /* vector along which asteroid moves */
	int    rot ;           /* current degrees of rotation */
	int    rot_delta ;     /* degress to rotate per unit time */
	vector rot_axis ;      /* axis asteroid rotates around */
} ;

typedef struct asteroid asteroid ;


struct torpedo {
	int in_flight_flag ;
	int is_tracer ;
	int trail_num ;
	float ttl ; 				/* time to live */
	vector pos ;
	vector pos_delta ;
	int bb_zrot ;   /* the number of degrees to rotate the torp billboard about the z axis */
} ;

typedef struct torpedo torpedo ;


struct torpedo_trail {
	int is_running_flag ;
	int is_ending_flag ;
	int trail_start ;
	int trail_end ;
	vector pos[TORP_TRAIL_LEN] ;
} ;

typedef struct torpedo_trail torpedo_trail ;


struct node {
	asteroid rock ;
	struct node *next ;
} ;

typedef struct node node ;


struct blast {
	int    is_active ;	
	float  ttl ;
	vector pos ;
	float  scalar ;
	int	   type ; /* which explosion display list to use */	
} ;

typedef struct blast blast ;


struct wind_part {
	vector pos ;  /* position on windshield glass */
	float  size ; /* size of impact circle */
	float  t_start ; /* time to start */
	float  t_end ;   /* time to end */
} ;

typedef struct wind_part wind_part ;


struct score_file_entry {
	char         name[MAX_TOKEN_LEN] ;
	unsigned int score ;
	char         date[MAX_TOKEN_LEN] ;
	int			 new_score_flag ; /* this value isn't saved to disk. */
} ;

typedef struct score_file_entry score_file_entry ;


struct keyboard_map {

	/* primary control keys */

	char end_game ;
	char pause ;
	
	/* movement keys */
	char slide_forward ;
	char slide_backward ;
	char slide_left ;
	char slide_right ;
	char slide_up ;
	char slide_down ;
	char stop ;
	char rot_aboutx_cw ;
	char rot_aboutx_ccw ;
	char rot_abouty_cw ;
	char rot_abouty_ccw ;
	char rot_aboutz_cw ;
	char rot_aboutz_ccw ;
	
	char headlight_toggle ;
	char target_asteroid ;
	char auto_target_asteroid ;
	char fire_torp ;
	
	/* feature toggles */
	char crosshair_toggle ;
	char draw_tracer_toggle ;
	char draw_shield_toggle ;
	char draw_dust_toggle ;
	char draw_torp_lensflare_toggle ;
	char asteroid_texture_toggle ;
	char nebula_texture_toggle ;
	char torp_texture_toggle ;
	char axes_toggle ;
	char play_sound_toggle ;
	char reverse_mouse_x_toggle ;
	char reverse_mouse_y_toggle ;
} ;

typedef struct keyboard_map keyboard_map ;


struct user_options_map {

	/* Flags for user preferences 0=off 1=on */
	int axes_flag ;                /* turn on and off the absolute x,y,z axes. I use this for debugging. */
	int crosshair_flag ;           /* turn the blue crosshair on or off */
	int draw_tracer_flag ;         /* should the torpedo tracer light be drawn */ 
	int draw_shield_flag ;         /* should the green alpha-blended shield effect be used */
	int draw_dust_flag ;           /* should the dust particles be drawn */ 
	int texture_flag ;             /* should the asteroids be textured */
	int nebula_texture_flag ;      /* should the nebula textures be used */
	int play_sound_flag ;          /* should sound be used */
	int torp_texture_flag ;        /* should the torps be textured */
	int draw_torp_lensflare_flag ; /* should the torp lens flares be drawn */

	/* more user preferences */
	float mouse_xdir ;			/* 1 = mouse movement turns the normal direction, -1 = mouse movement turns opposite direction (should be -1 or 1 only) */
	float mouse_ydir ;			/* 1 = mouse movement turns the normal direction, -1 = mouse movement turns opposite direction (should be -1 or 1 only) */
	int num_asteroids ;			/* number of asteroids, set by #define or command line */
	int windowx ; 		/* window x size, changed in reshape(). set by #define or on command line */
	int windowy ; 		/* window y size, changed in reshape(). set by #define or on command line */
	int window_xpos ; 	/* window left corner x offset */
	int window_ypos ; 	/* window left corner y offset */
} ;

typedef struct user_options_map user_options_map ;



/*
 *
 *  FUNCTION DECLARATIONS
 *
 */



/* defined in main.c */
void glut_initialization(int *argc, char **argv) ;
void gl_initialization(void) ;
void my_initialization(void) ;
void idle(void) ;
void keyboard_handler( unsigned char key, int x, int y ) ;
void mouse_handler( int button, int state, int x, int y) ;
void reshape(int w, int h) ;
void passive_mouse_handler(int x, int y) ;
void end_game(void) ;
void parse_command_line(int argc, char **argv) ;
void display_copyright_on_terminal(void) ;

/* defined in oglutils.c */
void draw_axes(void) ;
void apply_v2m( vector *v, float m[16] ) ;
float dot(vector u, vector v) ;
float length( float a, float b, float c) ;
int gen_rand_int(int min, int max) ;
void gen_rand_uv( vector *v ) ;
float gen_rand_float(float min, float max) ;
void draw_rectangle(void) ;
void find_rotation_angles(vector v, float *x, float *y) ;
int char2int(char c) ;
void find_rotation_mat(float angle, vector v, float m[16]) ;

/* view.c */
void reset_view(void) ;
void rotate_view_aboutx(float a) ;
void rotate_view_abouty(float a) ;
void rotate_view_aboutz(float a) ;

/* move.c */
void slide_alongx(float d) ;
void slide_alongy(float d) ;
void slide_alongz(float d) ;
void turn_aboutx(float t) ;
void turn_abouty(float t) ;
void turn_aboutz(float t) ;

/* display.c */
void display(void) ;
void draw_asteroids(void) ;
void draw_torps(void) ;
void draw_shield(void) ;
void draw_sun(void) ;
void draw_explosions(void) ;
void draw_hud(void) ;
void draw_mouse_pos(void) ;
void draw_dust(void) ;
void draw_nebula_billboards(void) ;
void draw_expl_winshield_particles(void) ;
void draw_torp_trails(void) ;

/* light.c */
void gl_init_light(void) ;
void toggle_headlight(void) ;
void update_headlight(void) ;
void gl_init_headlight(void) ;

/* thrust.c */
void init_thrust(void) ;
void process_motion(int value) ;
void apply_thrust(int value) ;
void stop(void) ;

/* asteroid.c */
void gen_rand_asteroid(asteroid *a) ;
void process_asteroid_motion(int value) ;
void init_asteroids(void) ;
node *add_node(void) ;
void print_list(void) ;
void init_asteroid_data(void) ;
void init_asteroid_dls(void) ;
void gen_rand_ast(int ps, int ts, float pert, int dl, float scalar) ;

/* torpedo.c */
void init_torpedo(void) ;
void fire_torpedo(void) ;
void process_torpedo_motion(int value) ;
void init_torps(void) ;
void init_torpedo_trails(void) ;
int get_free_torpedo_trail(void) ;
void process_torpedo_trails(void) ;
void calc_torp_billboard_rot(vector v, float *x, float *y) ;
void sort_torps(void) ;
int compare_torps(const void *pa, const void *pb) ;
void draw_torp_plasma(void) ;

/* collision.c */
void check_collision(node *previous) ;
float dist(vector v1, vector v2) ;
void apply_collision_v(vector *v, int mass) ;
void correct_positions(vector rock_v, vector *rock_p, float distance) ;
void process_deleted_ast(void) ;

/* shield.c */
void init_shield(void) ;
void process_shield(void) ;

/* hud.c */
void init_crosshair(void) ;
void draw_crosshair(void) ;
void draw_text_in_color( float x, float y, char *string, float r, float g, float b, float a) ;
void draw_shield_gauge(void) ;
void init_hud(void) ;
void init_collision_warning_gauge(void) ;
void draw_collision_warning_gauge(void) ;
void scan_for_collision_warning(asteroid a) ;
int draw_gauge(void) ;
void init_torp_gauge(void) ;
void draw_torp_gauge(void) ;

/* sun.c */
void init_sun(void) ;
void init_nebula_billboards(void) ;
void init_lens_flares(void) ;
void draw_lens_flares(void) ;

/* explosion.c */
void activate_explosion(vector pos) ;
void init_explosions(void) ;
void process_explosion(int value) ;
void init_explosion_data(void) ;
void init_explosion_dl(void) ;
void gen_rand_explosion_particles(int ps, int ts, float pert, int dl) ;
void init_explosion_winshield_particle_dl(void) ;

/* fps.c */
void init_fps(void) ;
float fps(void) ;

/* calc_frame.c */
void calculate_frame(void) ;
void capture_passive_mouse_input(void) ;

/* dust.c */
void init_dust(void) ;
void process_dust(void) ;

/* target.c */
void target_asteroid(void) ;
void init_target_dls(void) ;
void draw_lcs(asteroid a) ;
void draw_steering_cue(asteroid a) ;
int is_targeted(asteroid a) ;
void drop_target(void) ;
void draw_target_vector(asteroid a) ;
void draw_target_box(asteroid a) ;
void auto_target_asteroid(void) ;

/* split.c */
void add_new_asteroids( vector pos, int type) ;
void gen_rand_asteroid_at_pos(vector pos, asteroid *a, int type) ;

/* texture.c */
void init_textures(void) ;
char *read_bppm_tex(char *name, int h, int x, int y) ;
char *add_alpha_channel_to_RGB_image(char *image, int x, int y) ;
void change_alpha_for_RGBA_image(char *image, int x, int y, int r, int g, int b, int a) ;
void average_RGBA_alpha_and_floor(char *image, int x, int y, int floor) ;

/* wind_part.c */
void init_wind_parts(void) ;
void gen_rand_wind_part(wind_part *p, float secs) ;

/* pause.c */
int paused(void) ;
void start_pause(void) ;
void stop_pause(void) ;
void paused_display(void) ;
void paused_mouse_handler( int button, int state, int x, int y) ;
void paused_passive_mouse_handler(int x, int y) ;
void paused_keyboard_handler( unsigned char key, int x, int y ) ;


/* score.c */
void add_rock_to_score(asteroid a) ;
void print_score_on_terminal(void) ;
void read_to_eol(FILE *file) ;
int read_token(FILE *file, char *buf, int bufsize, char separator, char comment, char escape) ;
void read_score_file(void) ;
void write_empty_score_file(void) ;
int compare(const void *a, const void *b) ;
void sort_scores(void) ;
void write_score_file(void) ;
void init_score_display(void) ;
void score_keyboard_handler( unsigned char key, int x, int y ) ;
void score_display(void) ;

/* sound.c */
void init_sound(void) ;
char *read_unsigned_8bit_sound(char *name, int bytes) ;
void init_oss_sound_device(void) ;
void read_asteroid_sounds(void) ;
void play_sound(int type, int subtype) ;
void queue_sound(char *sound, int size) ;
void process_sound(void) ;

/* rcfile.c */
FILE *fopen_rcfile(void) ;
void read_rcfile(void) ;
void clean_token( char *buf, int bufsize ) ;
void write_rcfile_to_terminal(void) ;
int get_rcfile_token( FILE *file, char *buf, int bufsize ) ;
int is_token( char *buf1, char *buf2 ) ;
void key_char2string(char *buf, int bufsize, char character) ;
void opt_flag2string(char *buf, int bufsize, int flag) ;
void opt_mousedir2string(char *buf, int bufsize, int flag) ;
char key_string2char(char *buf) ;
int number_token(char *buf) ;

/* keyb.c */
void keyboard_handler( unsigned char key, int x, int y ) ;
void toggle_int_flag(int *flag) ;
void decay_keyb_rotx_scalar( void ) ;
void decay_keyb_roty_scalar( void ) ;
void decay_keyb_rotz_scalar( void ) ;
