/* Asteroids3D - a first person game of blowing up asteroids
 * Copyright (C) 2000 Stuart Mark Pomerantz
 *
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  Stuart Pomerantz
 *  3935 Stonecliffe drive
 *  Monroeville, PA 15146
 *  stuart@pitt.edu
 */

#include "asteroids3D.h"


/**
 ** These variables are ALL the global variables in the game 
 **/


/* Flags for internal program control */
int torps_in_flight_flag = 0 ;   /* are any torpedos in flight */
int mouse_roty_flag = 0 ;        /* set to 1 if the mouse has been moved out of the dead zone to rotate the ship about the y axis */
int mouse_rotx_flag = 0 ;        /* set to 1 if the mouse has been moved out of the dead zone to rotate the ship about the x axis */
int shield_activated_flag = 0 ;  /* has the shield been activated */
int explosions_active_flag = 0 ; /* are there explosions activated */
int paused_flag = 0 ;            /* is the game paused */
int game_end_flag = 0 ;          /* is the game over */
int tracer_light_flag = 0 ;      /* turn on the tracer light */
int keyb_rotx_scalar = 0 ; /* these are used for scaling keyboard rotation input */
int keyb_roty_scalar = 0 ;
int keyb_rotz_scalar = 0 ;


/* define the keyboard map */
/* beware the evils of adding a key and not correctly changing this definition */

keyboard_map keymap = {
	(char) END_GAME_KEY ,
	(char) PAUSE_KEY,

	(char) SLIDE_FORWARD_KEY ,
	(char) SLIDE_BACKWARD_KEY ,
	(char) SLIDE_LEFT_KEY ,
	(char) SLIDE_RIGHT_KEY ,
	(char) SLIDE_UP_KEY ,
	(char) SLIDE_DOWN_KEY ,
	(char) STOP_KEY ,
	(char) ROT_ABOUTX_CW_KEY ,
	(char) ROT_ABOUTX_CCW_KEY ,
	(char) ROT_ABOUTY_CW_KEY ,
	(char) ROT_ABOUTY_CCW_KEY ,
	(char) ROT_ABOUTZ_CW_KEY ,
	(char) ROT_ABOUTZ_CCW_KEY ,

	(char) HEADLIGHT_TOGGLE_KEY ,
	(char) TARGET_ASTEROID_KEY ,
	(char) AUTO_TARGET_ASTEROID_KEY ,
	(char) FIRE_TORP_KEY ,

	(char) CROSSHAIR_TOGGLE_KEY ,
	(char) DRAW_TRACER_TOGGLE_KEY ,
	(char) DRAW_SHIELD_TOGGLE_KEY ,
	(char) DRAW_DUST_TOGGLE_KEY ,
	(char) DRAW_TORP_LENSFLARE_TOGGLE_KEY ,
	(char) ASTEROID_TEXTURE_TOGGLE_KEY ,
	(char) NEBULA_TEXTURE_TOGGLE_KEY ,
	(char) TORP_TEXTURE_TOGGLE_KEY ,
	(char) AXES_TOGGLE_KEY ,
	(char) PLAY_SOUND_TOGGLE_KEY ,
	(char) REVERSE_MOUSE_X_TOGGLE_KEY ,
	(char) REVERSE_MOUSE_Y_TOGGLE_KEY 
} ;


user_options_map optmap = {
	(int) AXES_FLAG ,
	(int) CROSSHAIR_FLAG ,
	(int) DRAW_TRACER_FLAG ,
	(int) DRAW_SHIELD_FLAG ,
	(int) DRAW_DUST_FLAG ,
	(int) TEXTURE_FLAG ,
	(int) NEBULA_TEXTURE_FLAG ,
	(int) PLAY_SOUND_FLAG ,
	(int) TORP_TEXTURE_FLAG ,
	(int) DRAW_TORP_LENSFLARE_FLAG ,

	(float) MOUSE_XDIR ,
	(float) MOUSE_YDIR ,
	(int) NUM_ASTEROIDS ,
	(int) WINDOWX ,
	(int) WINDOWY ,
	(int) WINDOW_XPOS ,
	(int) WINDOW_YPOS
} ;






/* eye's local coordinate system. */
vector xaxis , yaxis , zaxis ;

/* eye's position in space */
vector pos ;

/* light0 ( the sun ) position,  
 * light1 (headlight) position & direction 
 * light2 torpedo tracer
 */
float l0_pos[4], l1_pos[4], l1_dir[4], l2_pos[4] ;

/* three important vectors: 
 *
 * 1) thrust 
 *
 * is the the vector which sums the thrust vectors
 * which the user has input but which have not yet been applied 
 * this is a "request"
 *
 * 2) velocity
 *
 * is the sum of the parts of the thrust vector which have been
 * applied (by the engine) and are now acting 
 * ( in a newtonian sense ) to move the ship.
 *
 * 3) rotation
 *
 * is the vector which holds the input requests for a rotation about an axis.
 * this is a "request" modulated by dt
 *
 */
vector thrust, velocity, rotation ;


/* define a torpedo array to hold, say, MAX_TORPS worth of them */
torpedo torp[MAX_TORPS] ;
torpedo sorted_torp[MAX_TORPS] ; /* doing this so alpha blending works correctly, this is scratch space */
int torp_dl ; /* torpedo display list */
int torp_tex ; /* torpedo billboard texture */
int torp_billboard_dl ; /* billboard for torpedo texture */

/* torpedo trails */
torpedo_trail torp_trail[MAX_TORP_TRAILS] ;

/* torp gauge dl 0 = off, 1 = on */
int torp_gauge_dl[2] ;

/* asteroid linked list head and tail pointers */
node *al_head, al_tail ;

/* shield display list, shield ttl, and shield strength */
int shield_dl ;
int shield_ttl ;
float shield_strength ;

/* crosshair display list */
int crossh_dl ;

/* sun display list */
int sun_dl ;

/* asteroid display list */
int ast_dl[NUM_AST_TYPES] ;

/* asteroid texture object list */
int ast_tex[NUM_AST_TYPES] ;

/* list of asteroids to be deleted each frame.
 * This list actually contains the node PRIOR to the node
 * to be deleted ( so I can hook PRIOR node to the deleted node's
 * "next" node. Hence the name. In any given frame, the largest
 * number of asteroids that can possibly be deleted is MAX_TORPS
 * so I make an array of this number of node pointers, just in case.
 * initialized in init_asteroids()
 */
node *delete_next_ast[MAX_TORPS] ;


/* explosion array, which is defined to be MAX_BLASTS > MAX_TORPS size since
 * there may be more than MAX_TORPS number of simultanious explosions
 * since the explosions (currently) last longer then the torp ttl.
 */
blast explosion[MAX_BLASTS] ;
int   expl_dl[NUM_BLAST_TYPES];
int   expl_winshield_particle_dl ;
int   expl_winshield_particle_count = 0 ;

/* track the mouse cursor */
int mouse_x, mouse_y ;

/* the proportion the mouse has moved out of the dead zone */
float mouse_xprop = 0 ;
float mouse_yprop = 0 ;

/* dust particle array */
vector dust_part[NUM_DUST_PART][NUM_CHILD_DUST_PART] ;

/* target box display list */
int target_box_dl, target_circle_dl ;
int target_arrow_on_dl, target_arrow_off_dl ;
int target_steering_cue_dl ;

/* a pointer to, well, that's pretty obvious */
node *currently_targeted_asteroid = NULL ;

/* nebula billboard texture object list, and nebula display list */
int nebula_tex[NUM_NEBULA_TEX] ;
int nebula_dl[NUM_NEBULA_TEX] ;

/* number of asteroids (for ship's "scanner" ) */
int num_asteroids_on_scanner ;


/* for collision warning gauge 0 = off, 1 = on */
int cwg_rear_dl[2] ; 
int cwg_above_dl[2], cwg_below_dl[2] ;
int cwg_left_dl[2],  cwg_right_dl[2] ;
int cwg_rear, cwg_above, cwg_below, cwg_left, cwg_right ; /* also 0 = off, 1 = on  */

/* windsheid particle array */
wind_part winshield_particle[MAX_BLASTS*NUM_WP_PER_BLAST] ;


/* last but not least, what would a game be without a score */
unsigned int score = 0 ;


/* time to end hud gauge flicker, intially off. Turned on after a collision. */
float flicker_hud_ttl = 0 ;


/* the number of game time ticks that have past */
unsigned long ticks = 0 ;

/* dt = delta time, the amount of time (in seconds)
 *  which passes with each game tick 
 */
float dt = 1.0/TPS ;

/* this is the number of ticks per second, which I may want to change dynamically */
float tps = TPS ;


/* frame calculation thread id, and the run_yield
 * semaphore for interleaving calculation and display updates
 */
pthread_t fc_tid ;
sem_t run_yield ;

 
/* OSS/Free audio device file descriptor */
int audio_device_fd ;

/* my sound buffer (doing this so sounds can be interrupted) */
char sound_buffer[MAX_SOUND_BUF_BYTES] ;
int  sound_buffer_start = 0, sound_buffer_end = 0 ;

/* sounds */
sound ast_snd[NUM_AST_TYPES] ; /* asteroid sounds */
sound ta_snd ; /* target aquired sound */
sound tf_snd ; /* torpedo fired sound */
sound coll_snd ; /* shield collision sound */


/* data structure to hold the scores read from the score file */
score_file_entry score_entry[NUM_SCORES] ;

/* lens flares */
lens_flare sun_flare[MAX_FLARES] ;
lens_flare torp_flare[MAX_FLARES] ;

/* the full pathname of the rcfile */
char *rcfile = NULL ;





int main(int argc, char **argv) {


	parse_command_line( argc, argv ) ;
	
	display_copyright_on_terminal() ;
	
	read_rcfile() ;
	
	glut_initialization(&argc, argv) ;
	gl_initialization() ;
	my_initialization() ;

	glutMainLoop();
	
	return(0) ;
}


void glut_initialization(int *argc, char **argv) {

	/* pass command line args to glut */
	glutInit(argc, argv);
	
	glutInitDisplayMode(GLUT_DOUBLE|GLUT_RGB|GLUT_DEPTH);
	
	glutInitWindowSize(optmap.windowx, optmap.windowy);
	glutInitWindowPosition(optmap.window_xpos, optmap.window_ypos);
	glutCreateWindow(COPYRIGHT);
	

	glutDisplayFunc( display );
	glutMouseFunc( mouse_handler ); 
	glutKeyboardFunc( keyboard_handler );
	glutIdleFunc( idle ) ;
	glutReshapeFunc( reshape ) ;
	glutPassiveMotionFunc( passive_mouse_handler ) ;
	

	/* glutTimerFunc( MSEC, check_collision, 1) ;	 */

	/* center the mouse cursor -- see glutPassiveMouseFunc */
	glutWarpPointer(optmap.windowx/2, optmap.windowy/2) ; 
	
	/* looks better than the arrow, I think */
	/* glutSetCursor(GLUT_CURSOR_CROSSHAIR) ;*/
	glutSetCursor(GLUT_CURSOR_NONE) ;
	
	return ;
}

void gl_initialization(void) {
		
	glClearColor(0.0, 0.0, 0.0, 0.0);
	 
	glMatrixMode(GL_PROJECTION); 
   	glLoadIdentity();
	glFrustum( LEFT, RIGHT, BOTTOM, TOP, ZNEAR, ZFAR ) ;
	 
	glEnable(GL_CULL_FACE) ;
	
	glEnable(GL_DEPTH_TEST);

	glEnable(GL_TEXTURE_2D) ;
		 
	return ;
}


void my_initialization(void) {

	printf("Initializing...\n");

	reset_view() ;
	gl_init_light() ;	
	gl_init_headlight() ; /* reset_view() must be called before gl_init_headlight */
	init_thrust() ;
	init_torps() ;
	init_textures();
	init_asteroids() ;
	init_shield()	 ;
	init_hud() ;
	init_sun() ; /* must be called after  gl_init_light */
	init_explosions();	
	init_dust() ;
	init_target_dls() ;
	init_nebula_billboards() ; /* must be called after  gl_init_light, and init_textures */
	init_wind_parts() ;
	init_sound() ;
	init_lens_flares() ;


	/* finally, and always last,
	 * set up the run_yield semaphore and
	 * crank up the frame calculation thread 
	 */
	
	sem_init( &run_yield, 0 , 1) ;

	pthread_create( &fc_tid, NULL, (void *)calculate_frame, NULL );
	
	printf("Done.\n");	
	
	return ;
}


void idle(void) {

	/* calculate_frame() is now a thread by itself, hope this works */
	/* calculate_frame() ;*/
		
	glutPostRedisplay() ;
	
	return ;
}


void reshape(int w, int h) {
	
	glViewport(0, 0, (GLsizei) w, (GLsizei) h);

	optmap.windowx = w ;
	optmap.windowy = h ;
	
	return ;
}



void mouse_handler( int button, int state, int x, int y) {

	/* mouse x,y are given in window coordinates
	 */


	switch( button ) {
	
		case GLUT_LEFT_BUTTON:
			if( state == GLUT_DOWN ) {
				fire_torpedo() ;
			}
		break ;


		case GLUT_RIGHT_BUTTON:
			if( state == GLUT_DOWN ) {
			}
		break ;
	}
	
	return ;
}

void passive_mouse_handler(int x, int y) {
	int left_bound   = optmap.windowx/2 - 50 ; /* dead zone boundaries */
	int right_bound  = optmap.windowx/2 + 50 ;
	int top_bound    = optmap.windowy/2 - 50 ;
	int bottom_bound = optmap.windowy/2 + 50 ;

	/* set the mouse position globals */
	mouse_x = x ;
	mouse_y = y ;

	
	/* if mouse is inside dead zone */
	if(  (x > left_bound) && (x < right_bound) &&  
		 (y > top_bound ) && (y < bottom_bound) ) {
		mouse_roty_flag = 0 ;
		mouse_rotx_flag = 0 ;
		return ;
	}

	
	/* if mouse is not in the dead zone set the flag */
	
	if( x > right_bound) {
		mouse_roty_flag = 1 ;
		mouse_yprop = (float)(x-right_bound)/(float)(optmap.windowx-right_bound) ;
	}
		

	if( x < left_bound) {
		mouse_roty_flag = 1 ;
		mouse_yprop = -1*(1 - (float)(x)/(float)left_bound) ;
	}

	
	if( y < top_bound) {
		mouse_rotx_flag = 1 ;
		mouse_xprop = (1 - (float)(y)/(float)top_bound) ;
	}


	if( y > bottom_bound) {
		mouse_rotx_flag = 1 ;	
		mouse_xprop = -1*(float)(y-bottom_bound)/(float)(optmap.windowy-bottom_bound)  ;		
	}

	/* for reversing mouse direction */
	
	mouse_yprop = mouse_yprop*optmap.mouse_ydir ;

	mouse_xprop = mouse_xprop*optmap.mouse_xdir ;
	
	return ;
}

void end_game(void) {

	game_end_flag = 1 ;  /* this causes the calculation thread to exit */
	init_score_display() ;

	return ;
}


void parse_command_line(int argc, char **argv) {
	int i ;
	int temp1, temp2, temp3, temp4 ;

	for( i = 0 ; i < argc ; i++ ) {

		if( strcmp("-help",argv[i]) == 0 ) {
			printf("valid command line options:\n") ;
			printf("   -num <number>\n"
			       "         to set the number of asteroids\n\n") ;
			printf("   -geometry <width>x<height>+<xoffset>+<yoffset>\n"
			       "         to set the geometry of the window (in pixels)\n\n") ;
			printf("   -rcfile <filename>\n"
			       "         to use <filename> as the .a3drc file instead of $HOME/.a3drc\n\n") ;
			printf("   -freshrc <filename>\n"
			       "         to print out a fresh .a3drc file\n\n") ;
			exit(0) ;
		}

		if( strcmp("-num",argv[i]) == 0 ) {
			if( i + 1 < argc ) {
				if( sscanf(argv[i+1],"%d\n",&temp1) == 1 ) {
					if( temp1 > 1 )
						optmap.num_asteroids = temp1 ;
					i++ ;
				}
			}
			continue ;
		}


		if( strcmp("-geometry",argv[i]) == 0 ) {
			if( i + 1 < argc ) {
				if( sscanf(argv[i+1],"%dx%d+%d+%d\n", &temp1, &temp2, &temp3, &temp4) >= 2 ) {			
					if( (temp1 > 0) && (temp2 >0) ) {
						optmap.windowx = temp1 ;
						optmap.windowy = temp2 ;
					}
					
					if( temp3 >= 0 )
						optmap.window_xpos = temp3 ;
						
					if( temp4 >= 0 )
						optmap.window_ypos = temp4 ;
				}
			}
		}


		if( strcmp("-rcfile",argv[i]) == 0 ) {
			if( i + 1 < argc ) {
				temp1 = strlen(argv[i+1]) + 1 ;
				rcfile = calloc( temp1, sizeof(char) ) ;
				memcpy(rcfile, argv[i+1], temp1 ) ;
			}
		}

		if( strcmp("-freshrc",argv[i]) == 0 ) {
			write_rcfile_to_terminal() ;
			exit(0) ;
		}


	} /* end for each argument */

	return ;
}

void display_copyright_on_terminal(void) {
	printf(COPYRIGHT) ;
	printf("Asteroids3D comes with ABSOLUTELY NO WARRANTY; for details\n"
           "see the file COPYRIGHT which should have come with this program.\n"
		   "This is free software, and you are welcome to redistribute it\n"
		   "under certain conditions; see the file COPYRIGHT\n"
		   "which should have come with this program for details.\n") ;
}
