/* Asteroids3D - a first person game of blowing up asteroids
 * Copyright (C) 2000 Stuart Mark Pomerantz
 *
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  Stuart Pomerantz
 *  3935 Stonecliffe drive
 *  Monroeville, PA 15146
 *  stuart@pitt.edu
 */
 
#include "asteroids3D.h"

/* find the rotation matrix for rotating 
 * angle DEGREES about vector V
 */
void find_rotation_mat(float angle, vector v, float m[16]) {
	float x,y,z ; /* direction cosines */
	float len ; /* length of the vector v*/
	float cos_a, sin_a ;
	
	/* calculate the normalized vector v and store its components in x,y,z */
	
	len = sqrt( v.x*v.x + v.y*v.y + v.z*v.z ) ;

	x = v.x/len ; y = v.y/len ; z = v.z/len ;

	
	
	/* pre-compute the cos(angle), sin(angle) */

	cos_a = cos( angle*PI/180.0 ) ;
	sin_a = sin( angle*PI/180.0 ) ;
	

	/* calculate the elements of the matrix */

	/* first row */
	
	m[0] = x*x + cos_a*( 1 - x*x )   ; 
	m[1] = x*y - cos_a*x*y - z*sin_a ; 
	m[2] = x*z - cos_a*x*z + y*sin_a ; 
	m[3] = 0 ; 
	

	/* second row */
	
	m[4] = y*x - cos_a*y*x + z*sin_a ; 
	m[5] = y*y + cos_a*( 1 - y*y )   ; 
	m[6] = y*z - cos_a*y*z - x*sin_a ; 
	m[7] = 0 ; 

	/* third row */
	
	m[8]  = z*x - cos_a*z*x - y*sin_a ; 
	m[9]  = z*y - cos_a*z*y + x*sin_a ; 
	m[10] = z*z + cos_a*( 1 - z*z )   ; 
	m[11] = 0 ; 

	/* fourth row */
	
	m[12] = 0 ; 
	m[13] = 0 ; 
	m[14] = 0 ; 
	m[15] = 1 ; 
	
	return ;
}


int char2int(char c) {
	int i ;
	int k = 0 ;
	int mask = 1 ;
	
	/* assuming a char is 8 bits */
	
	for( i = 0 ; i < 8 ; i++ ) {
	
		if( (c & mask) != 0 ) {
			k = k + (int) pow((double)2,(double)i) ;
		}
		mask = mask << 1 ;
	}

	return(k) ;
}


void find_rotation_angles(vector v, float *x, float *y) {
	float rho, phi, theta ;
	float cos_a ; /* cosine of the angle */
	vector i,j,k ; /* i,j,k are just for correcting the angle if v
	                  is not in the +x, +y, +z quadrant  */

	/* given a vector in rectangular co-ordinates,
	 * convert it to spherical coordinates.
	 * This will yield the vector's rotation about the
	 * y-axis (theta) and its rotation about the x-axis (phi)
	 *
	 * see schaum's calculus book on page 539 & 540
	 *
	 * since my coordinate system is layed out differently
	 * than schaums, my spherical coorinates look like this:
	 *
	 * x = p sin(phi) sin(theta)
	 * y = p cos(theta)
	 * z = p sin(phi) cos(theta)
	 *
	 * e.g. in my coordinate system, z comes out of the page, and
	 * x and y are in the plane of the page. In schaums, x comes out of
	 * the page and z and y are in the plane of the page.
	 *
	 */

	/*
	 * Note that after the rotations are computed,
	 * they must be applied like this:
	 *
	 *	glPushMatrix() ;
	 *		glRotatef(y,0,1,0) ;
	 *		glRotatef(x,1,0,0) ;
	 *
	 *      display the object
	 *
	 * 	glPopMatrix() ;
     */

	/* one last rub, I must make sure that 
	 * the number I'm feeding to acos is between -1 and 1
	 */


	/* get vector into spherical coordinates */

	
	rho = sqrt( dot(v,v) ) ;
	
	cos_a = v.y/rho ;
	
	if( cos_a < -1 )
		cos_a = -1 ;

	if( cos_a > 1 )
		cos_a = 1 ;	
	
	phi = acos( cos_a ) ;
	
	cos_a = v.z/(rho*sin(phi)) ;
	
	if( cos_a < -1 )
		cos_a = -1 ;

	if( cos_a > 1 )
		cos_a = 1 ;	
		
	theta = acos( cos_a ) ;


	phi = phi*180.0/PI ;
	
	theta = theta*180.0/PI ;

	*y = theta ;
	
	*x = -(90 - phi) ;


	/* the subroutine could return here, except that I have 
	 * to account for all 8 octants, I have to do some checking
	 */

	/* set up the unit basis vectors */
	
	i.x = 1 ; i.y = 0 ; i.z = 0 ; i.w = 1 ;
	j.x = 0 ; j.y = 1 ; j.z = 0 ; j.w = 1 ;
	k.x = 0 ; k.y = 0 ; k.z = 1 ; k.w = 1 ;


	if( dot(v,i) < 0 )
		*x = 180 - *x ;

	if( dot(v,i) < 0 )
		*y = 180 - *y ;
	
	return ;
}




float length( float a, float b, float c) {

	return( sqrt( a*a + b*b + c*c ) ) ;
}



void gen_rand_uv( vector *v ) {
	int x,y,z ;
	float m[16] ;
	vector u, a ;

	/* initial unit vector down the z - axis */		
	u.x = 0 ; u.y = 0 ; u.z = 1 ;
	
	/* srand(time(NULL)) ; */
	
	x = 0 +(int) (359.0*rand()/(RAND_MAX+1.0)) ;
	y = 0 +(int) (359.0*rand()/(RAND_MAX+1.0)) ;
	z = 0 +(int) (359.0*rand()/(RAND_MAX+1.0)) ;

	/* set the axis a to be the x axis first */
	
	a.x = 1 ; a.y = 0 ; a.z = 0 ; a.w = 1 ;

	/* rotate u */
	find_rotation_mat( x, a, m ) ;
	apply_v2m( &u,m ) ;
	

	/* set the axis a to be the y axis next */
	
	a.x = 0 ; a.y = 1 ; a.z = 0 ; a.w = 1 ;

	/* rotate u */
	find_rotation_mat( y, a, m ) ;
	apply_v2m( &u,m ) ;


	/* set the axis a to be the z axis finally */
	
	a.x = 0 ; a.y = 0 ; a.z = 1 ; a.w = 1 ;

	/* rotate u */
	find_rotation_mat( z, a, m ) ;
	apply_v2m( &u,m ) ;

	
	/* put the results back into v */	
	v->x = u.x ; v->y = u.y ; v->z = u.z ;

	return ;
}


/* generate random floats in .1 (tenth) increments */
float gen_rand_float(float min, float max) {
	float len = max - min ;
	float k ;

	/* srand(time(NULL)) ; */
		
	/* generate a number between 0 and len*10 */
	k = 0 + ( (10.0*len) *rand()/(RAND_MAX+1.0)) ;

	/* scale the number back down */
	k = k/10.0 - (float)abs(min) ;	
	
	return(k) ;
}



int gen_rand_int(int min, int max) {
	int len = max - min + 1 ;
	int k ;

	/* srand(time(NULL)) ; */	
		
	/* generate a number between 0 and len */
	k = 0 + (int) ( (float)len *rand()/(RAND_MAX+1.0)) ;

	/* scale the number back down */
	k = k - abs(min) ;	
	
	/* printf("min %d, max %d, picked %d\n",min,max,k) ;*/
	
	return(k) ;
}



float dot(vector u, vector v) {
	float dot ;

	dot = u.x*v.x + u.y*v.y + u.z*v.z ;
	
	return(dot) ;
}


void apply_v2m( vector *v, float m[16] ) {
	vector a ;
	
	a.x = v->x ; a.y = v->y ; a.z = v->z ; a.w = v->w ;
	
	v->x = a.x*m[0] + a.y*m[1] + a.z*m[2]  + a.w*m[3] ;
	v->y = a.x*m[4] + a.y*m[5] + a.z*m[6]  + a.w*m[7] ;
	v->z = a.x*m[8] + a.y*m[9] + a.z*m[10] + a.w*m[11] ;


	return ;
}



void draw_axes(void) {
	float high_red[]  = { 1.0, 0.0, 0.0, 1.0} ;
	float low_red[]   = { 0.3, 0.0, 0.0, 1.0} ;
	float high_green[]= { 0.0, 1.0, 0.0, 1.0} ;
	float low_green[] = { 0.0, 0.3, 0.0, 1.0} ;
	float high_blue[] = { 0.0, 0.0, 1.0, 1.0} ;
	float low_blue[]  = { 0.0, 0.0, 0.3, 1.0} ;
	float white[]     =  {1.0, 1.0, 1.0, 1.0} ;
	float ad[] = { 0, 0, 0, 1 } ; 
	float k = 5.0 ;
	
	glLineWidth( 3.0 ) ;
	glPointSize( 6.0 ) ;
			
	glBegin( GL_LINES ) ;

		glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, ad) ;

			/* x axis RED */
		glMaterialfv(GL_FRONT, GL_EMISSION, high_red) ;
		glVertex3f( k , 0.0 , 0.0 ) ;
		glVertex3f( 0.0 , 0.0 , 0.0 ) ;
                
		glMaterialfv(GL_FRONT, GL_EMISSION, low_red) ;
		glVertex3f( -k , 0.0 , 0.0 ) ;
		glVertex3f( 0.0 , 0.0 , 0.0 ) ;
                
			/* y axis GREEN */
		glMaterialfv(GL_FRONT, GL_EMISSION, high_green) ;
		glVertex3f( 0.0 , k ,0.0 ) ;
		glVertex3f( 0.0 , 0.0 , 0.0 ) ;
                                
		glMaterialfv(GL_FRONT, GL_EMISSION, low_green) ;
		glVertex3f( 0.0 , -k , 0.0 ) ;
		glVertex3f( 0.0 , 0.0 , 0.0 ) ;
	
			/* z axis BLUE */
		glMaterialfv(GL_FRONT, GL_EMISSION, high_blue) ;
		glVertex3f( 0.0 , 0.0 , k ) ;
		glVertex3f( 0.0 , 0.0 , 0.0 ) ;
                                
		glMaterialfv(GL_FRONT, GL_EMISSION, low_blue) ;
		glVertex3f( 0.0 , 0.0 , -k ) ;
		glVertex3f( 0.0 , 0.0 , 0.0 ) ;
	glEnd() ;


	glBegin( GL_POINTS ) ;
		glMaterialfv(GL_FRONT, GL_EMISSION, white) ;	
		glVertex3f( k , 0.0 , 0.0 ) ;
		glVertex3f( 0.0 , k ,0.0 ) ;
		glVertex3f( 0.0 , 0.0 , k ) ;
	glEnd() ;
	
	glLineWidth( 1.0 ) ;
	glPointSize( 1.0 ) ;
	
}


void draw_rectangle(void) {
	float high_blue[] = { 0.0, 0.0, 1.0, 1.0} ;
	float k = 3.0 ;
	
	glLineWidth( 2.0 ) ;
	glMaterialfv(GL_FRONT, GL_EMISSION, high_blue) ;	
	
	glBegin(GL_LINE_LOOP) ;
	
		/* upper left corner */
		glVertex3f( -k/2, -k/2, 0 ) ;
	
		/* upper right corner */
		glVertex3f(  k/2, -k/2, 0 ) ;

		/* lower right corner */
		glVertex3f(  k/2, k/2, 0 ) ;
	
		/* lower left corner */
		glVertex3f( -k/2, k/2, 0 ) ;
	
	glEnd() ;

	
	glBegin(GL_LINES) ;
		glVertex3f( 0, 0, 0 ) ;
		glVertex3f( 0, 0, k ) ;
	glEnd() ;
	
	return ;
}
